# This is the left section part.
from flet import Page, AnimatedSwitcher, Container, Row, Column
import flet
import time

from ..widgets.All import all_widgets
from ..ui_toolkit.widget_browser_frame import Widget_Browse_Frame

class leftSection:
    def __init__(self, page:Page, main_class, main_row:Row) -> None:
        self.page : Page = page
        self.main_class = main_class
        self.main_row = main_row

        self.self_ui = AnimatedSwitcher(transition=flet.AnimatedSwitcherTransition.FADE, duration=500, reverse_duration=250, switch_in_curve=flet.AnimationCurve.EASE, switch_out_curve=flet.AnimationCurve.EASE_IN_BACK)
        self.main_container = Container(bgcolor=page.bgcolor, width=page.width/4-30, height=page.height)
        self.self_ui.content = self.main_container
        main_row.controls.append(self.self_ui)

        #? Show all widgets.
        self.show_all_widgets()


    def show_new_content (self, container:Container):
        self.main_container = container
        self.self_ui.content = container

        if self.self_ui.page != None:
            self.self_ui.update()
        else:
            self.page.update()
        
        if hasattr(self.main_class, "on_page_resize"):
            try: self.main_class.on_page_resize()
            except: pass
    

    def show_all_widgets (self):
        page = self.page
        #* This is the default section.
        
        col = Column(scroll=True)
        title = flet.Text("Widgets", size=25, weight="bold", color="white")
        col.controls.append(title)
        for wid in all_widgets:
            w = Widget_Browse_Frame(wid, all_widgets[wid], self.main_class.add_new_widget, self.main_class.current_page_name)
            col.controls.append(w)
        

        self.main_container = Container(bgcolor=page.bgcolor, width=page.width/4, height=page.height)
        self.main_container.content = col
        self.show_new_content(self.main_container)
        self.page.update()
    

    def show_the_suggestions (self):
        pass


import flet


class Image(object):
    def __init__(self, main_class, parent, *args, **kwargs) -> None:
        self.parent = parent
        self.main_class = main_class
        self.self_object = flet.Image(fit=flet.ImageFit.COVER)

        # all args
        self.args = {
            "src": {"type": str, "default_value": "https://picsum.photos/id/237/200/300"},
            "width": {"type": int, "default_value": 150},
            "height": {"type": int, "default_value": 150},
            "border_radius": {"type": int, "default_value": 75},
            "alignment": {"type": list, "options": ["left", "center", "right"], "default_value": "center"},
        }

        # Template dict
        # This is where the widget data will be stored.
        self.template = {
            "widget_class_name": "Image",
            "properties": {}
        }
        for p in self.args:
            self.template["properties"][p] = self.args[p]["default_value"]

        self.update()

    def update(self, new_props: dict = None):
        img = self.self_object
        props = self.template["properties"]

        if new_props is not None:
            for i in new_props:
                self.template["properties"][i] = new_props[i]

        img.src = props["src"]
        img.width = props["width"]
        img.height = props["height"]
        img.border_radius = props["border_radius"]

        if self.self_object.page is not None:
            self.self_object.update()

    def return_widget(self):
        props = self.template["properties"]
        if props["alignment"] == "left":
            return flet.Row([flet.Text("    "), self.self_object])
        elif props["alignment"] == "center":
            return flet.Row([self.self_object], alignment=flet.MainAxisAlignment.CENTER)
        else:
            return flet.Row([self.self_object, flet.Text("    ")], alignment=flet.alignment.center_right)



import flet
import time
import os
import json
import requests

from ..tools.create_storyboard import Create_StoryBoard


class CreateNewFile:
    def __init__(self, page: flet.Page, manage_class):
        page.window_resizable = False
        page.update()
        page.window_center()

        # Set page prop
        page.bgcolor = flet.colors.BLACK

        # Setup Content's mother.
        self.manage_class = manage_class
        self.page = page
        self.mother = flet.AnimatedSwitcher(
            expand=True,
            content=flet.Text(""),
            transition=flet.AnimatedSwitcherTransition.FADE,
            duration=500,
            reverse_duration=250,
            switch_in_curve=flet.AnimationCurve.EASE_IN
        )
        page.add(self.mother)

        time.sleep(0.3)
        # Run the first page.
        self.page_one()

    def page_one(self):
        page = self.page
        mother = self.mother
        main_column = flet.Column()
        mother.content = main_column

        main_column.controls.append(flet.Text("\n\n"))

        title_1 = flet.Text(
            "\n        Let's imagine 😃,\n        What if there is a ...",
            color=flet.colors.WHITE,
            size=36,
            weight=flet.FontWeight.BOLD
        )
        describe = flet.Text(
            "                  Everything start from the imagine skill.\n                  With Flet_StoryBoard you will make the impossible ✨.\n",
            color=flet.colors.WHITE60,
            size=15
        )
        next_button = flet.Container(
            flet.Row(
                [flet.Text("Start", color=flet.colors.BLACK)],
                alignment=flet.MainAxisAlignment.CENTER
            ),
            bgcolor=flet.colors.WHITE,
            width=100,
            height=35,
            border_radius=11,
            on_click=self.page_two
        )

        main_column.controls.append(title_1)
        main_column.controls.append(describe)
        main_column.controls.append(
            flet.Column(
                [
                    flet.Row(
                        [next_button],
                        alignment=flet.MainAxisAlignment.END
                    )
                ],
                alignment=flet.MainAxisAlignment.END,
                expand=True
            )
        )

        page.update()

    def page_two(self, *args):
        def on_type(event):
            if event.control.value == "":
                next_button.tooltip = "You must pick a name."
                next_button.disabled = True
                next_button.update()
            else:
                next_button.tooltip = "Click to generate the file."
                next_button.disabled = False
                page.title = "Flet StoryBoard - " + str(event.control.value)
                if os.path.isfile(f"{page.title}.fletsb"):
                    page.title = page.title + " | Warning: File name exist"
                page.update()
                self.file_name = str(event.control.value)

        page = self.page
        mother = self.mother
        main_column = flet.Column()
        mother.content = main_column

        main_column.controls.append(flet.Text("\n\n"))

        Title1 = flet.Text("\n        Hey sir 🎩,\n        What is the name ?", color=flet.colors.WHITE, size=36, weight="bold")
        main_column.controls.append(Title1)

        describe = flet.Text(
            "                  Lets choose a name of our UI.\n                  This name is the same as StoryBoad file name!.\n",
            color=flet.colors.WHITE60, size=15)
        main_column.controls.append(describe)

        StoryBoard_Name = flet.TextField(label="UI name", border_color=flet.colors.BLACK, cursor_color=flet.colors.BLACK,
                                         bgcolor=flet.colors.WHITE, border_radius=19, width=350, height=60, on_change=on_type,
                                         color=flet.colors.BLACK, on_submit=self.page_three)
        main_column.controls.append(flet.Row([StoryBoard_Name], alignment="center"))

        next_button = flet.Container(flet.Row([flet.Text("Generate", color=flet.colors.BLACK)], alignment="center"),
                                     bgcolor=flet.colors.WHITE, width=100, height=35, border_radius=11, on_click=self.page_three,
                                     disabled=True)
        main_column.controls.append(
            flet.Column([flet.Row([next_button], alignment="END")], alignment="END", expand=True))

        mother.update()
        page.update()

    def page_three(self, *args):
        page = self.page
        page.title = page.title
        mother = self.mother
        main_column = flet.Column()
        mother.content = main_column

        main_column.controls.append(flet.Text("\n\n"))

        Title1 = flet.Text("\n        Now🪄,\n        Just choose a template!!", color=flet.colors.WHITE, size=36, weight="bold")
        main_column.controls.append(Title1)

        describe = flet.Text(
            "                  Ok, what is a template ?.\n                  A template is a theme and pre-set page that make it quick to start.\n",
            color=flet.colors.WHITE60, size=15)
        main_column.controls.append(describe)

        r = flet.Row([
            self.make_a_template_button(flet.colors.WHITE24, "Blank new - default", "ADD_ROUNDED"),
            self.make_a_template_button(flet.colors.WHITE24, "Import template - soon", "GET_APP_OUTLINED", disable=True)
        ], alignment="center", spacing=15)
        main_column.controls.append(r)

        mother.update()
        page.update()

    def page_four(self, *args):
        page = self.page
        page.title = page.title
        mother = self.mother
        main_column = flet.Column()
        mother.content = main_column

        main_column.controls.append(flet.Text("\n\n"))

        Title1 = flet.Text("\n        Wait😄,\n        Building things for you.", color=flet.colors.WHITE, size=36, weight="bold")
        main_column.controls.append(Title1)

        mother.update()
        page.update()
        time.sleep(1)
        self.apply_suggestion_page()

    def apply_suggestion_page(self, *a):
        def dont_apply(e):
            Create_StoryBoard(self.file_name, "default", False)
            self.manage_class.file_name = self.file_name
            self.page.window_close()

        def apply_it(e):
            self.final_page_with_storyboard_suggest()

        page = self.page
        page.vertical_alignment = flet.MainAxisAlignment.CENTER
        page.title = page.title
        mother = self.mother
        main_column = flet.Column()
        mother.content = main_column
        col = main_column

        col.controls.append(flet.Row([flet.Text("", height=15)], alignment="center"))

        image = flet.Image(src="https://raw.githubusercontent.com/SKbarbon/Flet_StoryBoard/main/assets/IMG_3260.PNG",
                           width=250, height=250)
        col.controls.append(flet.Row([image], alignment="center"))

        title = flet.Text("Apply StoryBoard suggestions", size=20, weight="bold", color=flet.colors.WHITE)
        col.controls.append(flet.Row([title], alignment="center"))

        describe = flet.Text("""
        These are suggestions based on smart rules that help 
        you build your interface by suggesting 
        widgets based on your goal.
        """, size=13, color=flet.colors.WHITE)
        col.controls.append(flet.Row([describe], alignment="center"))

        apply_button = flet.ElevatedButton("Apply", bgcolor=flet.colors.WHITE, color=flet.colors.BLACK, width=250, on_click=apply_it)
        col.controls.append(flet.Row([apply_button], alignment="center"))

        no_btn = flet.TextButton(content=flet.Text("no thanks", color="red"), on_click=dont_apply)
        col.controls.append(flet.Row([no_btn], alignment="center"))

        mother.update()
        page.update()

    def final_page_with_storyboard_suggest(self):
        def apply_and_start(rule_name, rule_install_url):
            rules_file_content = requests.get(rule_install_url).text
            if not os.path.isdir("rules"):
                os.mkdir("rules")
            open(f"rules/{rule_name}.json", "w+", encoding="utf-8").write(str(rules_file_content))
            Create_StoryBoard(self.file_name, "default", True, rule_name)
            self.manage_class.file_name = self.file_name
            self.page.window_close()

        # ------
        page = self.page
        page.title = page.title
        mother = self.mother
        main_column = flet.Column()
        mother.content = main_column
        col = main_column
        col.scroll = True

        main_column.controls.append(flet.Text("\n\n"))

        Title1 = flet.Text("Choose your case.", color=flet.colors.WHITE, size=36, weight="bold")
        main_column.controls.append(flet.Row([Title1], alignment="center"))
        mother.update()
        describe = flet.Text("""
Choose your 'main' page goal from the 
options down bellow. This will install the right suggestions rules
for your main page case.
        """, size=13, color=flet.colors.WHITE)
        col.controls.append(flet.Row([describe], alignment="center"))
        mother.update()

        all_rules = requests.get("https://raw.githubusercontent.com/SKbarbon/Flet_StoryBoard/main/rules/all.json").text
        all_rules = json.loads(all_rules)

        for rule_option in all_rules["all_rules"]:
            col.controls.append(
                self.make_a_rule_button_option(rule_option, all_rules["all_rules"][rule_option], apply_and_start))

        col.scroll = True
        mother.update()

    def make_a_template_button(self, bgcolor, name, icon, disable=False):
        def on_hov(event):
            if event.data == "true":
                event.control.bgcolor = flet.colors.WHITE12
            elif event.data == "false":
                event.control.bgcolor = flet.colors.WHITE24
                event.control.update()
            event.control.update()

        cc = flet.Column(alignment="center", disabled=disable)
        c = flet.Container(width=220, height=140, bgcolor=bgcolor, border_radius=10, on_click=self.page_four,
                           on_hover=on_hov)
        cc.controls.append(c)

        Icon = flet.Icon(icon, size=20, color=flet.colors.WHITE)
        c.content = flet.Row([Icon], alignment="center")

        name_it = flet.Text(name, color=flet.colors.WHITE, size=13, text_align="center")
        cc.controls.append(flet.Row([name_it], alignment="center"))

        return cc

    def make_a_rule_button_option(self, name, url, on_click):
        def on_choose(e):
            on_click(name, url)

        b = flet.ElevatedButton(f"{name}", on_click=on_choose, bgcolor=flet.colors.WHITE, color=flet.colors.BLACK, width=150, height=45)

        return flet.Row([b], alignment="center")



import flet

from ...tools.color_picker import ColorPicker


def page_settings_page(settings_class):
    def allow_scrolling(e):
        settings_class.main_class.dict_content["storyboard_settings"]["allow_scroll"] = e.control.value
        page: flet.Page = settings_class.page
        page.show_snack_bar(
            flet.SnackBar(
                flet.Text(f"Done! But this is not saved, you must go click 'Save' on the editor."),
                open=True
            )
        )
        settings_class.main_class.preview_section.update_preview()
        page.update()

    def on_change_bgcolor(color):
        settings_class.main_class.dict_content["pages"][current_page_name]["settings"]["bgcolor"] = color
        page: flet.Page = settings_class.page
        page.show_snack_bar(
            flet.SnackBar(
                flet.Text(f"Done! But this is not saved, you must go click 'Save' on the editor."),
                open=True
            )
        )
        settings_class.main_class.preview_section.update_preview()
        page.update()

    v = settings_class.page_viewing_section
    v.clean()

    current_page_name = settings_class.main_class.current_page_name
    title = flet.Text(
        f"\n   Pages - {current_page_name}",
        color=flet.colors.WHITE,
        weight=flet.FontWeight.BOLD,
        size=28
    )
    v.controls.append(title)

    if "allow_scroll" not in settings_class.main_class.dict_content["storyboard_settings"]:
        settings_class.main_class.dict_content["storyboard_settings"]["allow_scroll"] = False

    row_allow_scrolling = flet.Row(
        [
            flet.Text("Allow scrolling", color="white", width=300),
            flet.Switch(
                on_change=allow_scrolling,
                value=settings_class.main_class.dict_content["storyboard_settings"]["allow_scroll"]
            )
        ],
        alignment=flet.MainAxisAlignment.CENTER
    )
    v.controls.append(row_allow_scrolling)

    cp = ColorPicker(v, settings_class.main_class.dict_content["pages"][current_page_name]["settings"]["bgcolor"],
                     on_choose_color=on_change_bgcolor, add_it=False, title_name="bgcolor")
    v.controls.append(
        flet.Container(
            flet.Row(
                [cp.v],
                alignment=flet.MainAxisAlignment.CENTER
            ),
            bgcolor=flet.colors.BLACK
        )
    )

    v.update()



from flet import Page, Container
import flet

from ..widgets.All import all_widgets


class viewerEngine:
    def __init__ (self, main_class, content_dict:dict, page_name, parent_view, widgets_parent_view, development=True):
        """
        This is the engine of preview.
        parent_view -> is the Container in the develop case, and Page in the production case
        widgets_parent_view -> is the Column in the develop case, and Page in the production case
        """
        self.main_class = main_class
        self.content = content_dict
        self.page_name = page_name
        self.parent_view = parent_view
        self.widgets_parent_view = widgets_parent_view
        self.development = development

        self.last_clicked = None # last widget clicked to edit. this is for make a border color around it to know its selected.

        self.update_page()
        self.push_views()

    def push_views (self):
        page_name : str = self.page_name
        content = self.content["pages"][page_name]
        sub_widgets = content["widgets"]

        self.widgets_parent_view.controls.clear()
        num = 0
        for widget in sub_widgets:
            if widget["widget_class_name"] in all_widgets:
                widget_class = all_widgets[widget["widget_class_name"]]["class"]
                widget_class = widget_class(self.main_class, self.widgets_parent_view, widget_number=num)
                if hasattr(widget_class, "support_sub_widgets"):
                    widget_class.update(widget["properties"], widget["widgets"])
                else:
                    widget_class.update(widget["properties"])
                if self.development:
                    self.create_development_container(widget_class.return_widget(), num)
                else:
                    self.widgets_parent_view.controls.append(widget_class.return_widget())
            num = num + 1

    
    def update_page (self):
        p : flet.Container = self.parent_view
        page_settings = self.content["pages"][self.page_name]["settings"]
        bgcolor = page_settings["bgcolor"]
        p.bgcolor = bgcolor

        if "allow_scroll" in self.main_class.dict_content["storyboard_settings"]:
            self.widgets_parent_view.scroll = self.main_class.dict_content["storyboard_settings"]["allow_scroll"]
    
    def create_development_container (self, cls, widget_number):
        def on_click (cls):
            if self.last_clicked != None:
                self.last_clicked.border = None
            self.last_clicked = c
            self.last_clicked.border = flet.border.all(2, flet.colors.BLUE_100)
            #? This function will call the main_page edit.
            self.main_class.edit_a_widget(widget_number)
        
        c = flet.Container(cls, on_click=on_click, border_radius=8)
        self.widgets_parent_view.controls.append(c)
        return c


import flet
from ...ui_toolkit.widget_browser_frame import Widget_Browse_Frame


class Row(object):
    def __init__(self, main_class, parent, *args, **kwargs) -> None:
        self.parent = parent
        self.main_class = main_class
        self.self_object = flet.Row()

        self.widget_number = kwargs["widget_number"]

        # Row Special args.
        self.dont_border_subs = True
        self.last_added = None
        self.support_sub_widgets = True

        if main_class.development_mode:
            self.development_mode = True
        else:
            self.development_mode = False

        # all args
        self.args = {
            "width": {"type": int, "default_value": 23},
            "height": {"type": int, "default_value": 150},
            "hide": {"type": bool, "default_value": False},
            "scroll": {"type": bool, "default_value": False},
            "auto_scroll_to_end": {"type": bool, "default_value": True},
            "expand": {"type": bool, "default_value": True},
            "alignment": {"type": list, "options": ["left", "center", "right"], "default_value": "center"},
            "sub_widgets_alignment": {"type": list, "options": ["left", "center", "right"], "default_value": "center"},
        }

        # Template dict
        # This is where the widget data will be stored.
        self.template = {
            "widget_class_name": "Row",
            "properties": {},
            "widgets": []
        }
        for p in self.args:
            self.template["properties"][p] = self.args[p]["default_value"]

        self.update()

    def update(self, new_props: dict = None, new_sub_widgets: dict = None):
        r = self.self_object
        props = self.template["properties"]

        if new_props is not None:
            for i in new_props:
                self.template["properties"][i] = new_props[i]

        if new_sub_widgets is not None:
            for w in new_sub_widgets:
                self.template["widgets"].append(w)

        r.width = props["width"]
        r.height = props["height"]
        r.visible = props["hide"] == False
        r.expand = props["expand"]
        r.scroll = props["scroll"]
        r.alignment = props["sub_widgets_alignment"]
        r.auto_scroll = props["auto_scroll_to_end"]

        if self.self_object.page is not None:
            self.self_object.update()

        self.update_preview()

    # Row tools
    def widgets_to_add_in(self, *args):
        """To show all widgets that are available to add to the row."""
        for i in args:
            i.control.visible = False
            i.control.update()

        def go_back(*e):
            self.main_class.left_section.show_new_content(copy_of_last_container)
            if self.last_added is not None:
                if self.last_added.page is None:
                    self.last_added.border = None
                    self.main_class.page.update()
                else:
                    self.last_added.border = None
                    self.last_added.update()
            self.main_class.on_page_resize()
            for i in args:
                i.control.visible = True
                i.control.update()

        page = self.main_class.page
        # This is the default section.

        # issue: scroll has not attribute True
        col = flet.Column(scroll=True)
        col.controls.append(flet.TextButton("< Back", on_click=go_back))
        title = flet.Text("Widgets in row", size=25, weight=flet.FontWeight.BOLD, color="white")
        col.controls.append(title)
        all_widgets = self.main_class.all_widgets
        for wid in all_widgets:
            w = Widget_Browse_Frame(wid, all_widgets[wid], self.add_new_widget, self.main_class.current_page_name)
            col.controls.append(w)

        main_container = flet.Container(
            content=col,
            bgcolor=page.bgcolor,
            width=page.width / 4,
            height=page.height
        )

        copy_of_last_container = self.main_class.left_section.self_ui.content
        self.main_class.left_section.show_new_content(main_container)
        self.main_class.on_page_resize()
        self.main_class.page.update()

    def add_new_widget(self, widget_name, page_name):
        """To add the widget, and open the edit to edit this new sub-widget."""
        self.dont_border_subs = False
        all_widgets = self.main_class.all_widgets
        widget_class = all_widgets[widget_name]["class"]
        widget_class = widget_class(
            self, self.preview_section,
            widget_number=len(self.template["widgets"]) - 1,
            is_a_sub=True
        )

        self.template["widgets"].append(widget_class.template)
        self.main_class.dict_content["pages"][self.main_class.current_page_name]["widgets"][self.widget_number]["widgets"] = self.template["widgets"]

        self.update_preview()
        editSubWidgetsEngine = self.main_class._editSubWidgetsEngine
        editSubWidgetsEngine = editSubWidgetsEngine(
            self.main_class,
            self,
            self.main_class.edit_section.main_column,
            len(self.template["widgets"]) - 1
        )
        self.main_class.page.update()

    def update_preview(self):
        all_widgets = self.main_class.all_widgets
        self.preview_section.controls.clear()
        sub_widgets = self.template["widgets"]
        num = 0

        for widget in sub_widgets:
            if widget["widget_class_name"] in all_widgets:
                widget_class = all_widgets[widget["widget_class_name"]]["class"]
                widget_class = widget_class(self.main_class, self.preview_section)
                if hasattr(widget_class, "support_sub_widgets"):
                    widget_class.update(widget["properties"], widget["widgets"])
                else:
                    widget_class.update(widget["properties"])
                if self.development_mode:
                    self.create_development_container(widget_class.return_widget(), num)
                else:
                    self.preview_section.controls.append(widget_class.return_widget())
            num = num + 1

        self.main_class.page.update()

    @property
    def preview_section(self):
        return self.self_object

    def return_widget(self):
        props = self.template["properties"]

        if props["alignment"] == "left":
            return flet.Row([flet.Text("    "), self.self_object])
        elif props["alignment"] == "center":
            return flet.Row([self.self_object], alignment=flet.MainAxisAlignment.CENTER, spacing=0)
        else:
            return flet.Row([self.self_object, flet.Text("    ")], alignment=flet.MainAxisAlignment.END)

    def create_development_container(self, cls, widget_number):
        if self.last_added is None:
            c = flet.Container(cls, border_radius=8, border=flet.border.all(2, flet.colors.YELLOW_300))
            self.last_added = c
        else:
            self.last_added.border = None
            c = flet.Container(cls, border_radius=8, border=flet.border.all(2, flet.colors.YELLOW_300))
            self.last_added = c

        self.preview_section.controls.append(c)
        return c



import flet


class ListPopup:
    def __init__(self, the_list=None, main_class=None, default_choice=None, title=None):
        if the_list is None:
            return

        self.title = title
        self.page: flet.Page = main_class.page
        self.main_class = main_class
        self.self_ui = flet.TextButton(f"{title}: f{default_choice}", on_click=self.show_popup)
        self.the_list = the_list
        self.__selected = default_choice

    def show_popup(self, *args):
        def on_search(e):
            for i in self.all_list_choices:
                col.controls.remove(i)

            self.all_list_choices = []

            if search_field.value == "":
                for icon in self.the_list:
                    name = str(icon)
                    if name.startswith("__"):
                        pass
                    else:
                        self.all_list_choices.append(self.generate_new(name, col))
                col.update()
                return

            self.all_list_choices = []
            self.all_list_choices.clear()
            s = str(search_field.value).lower()

            for i in self.the_list:
                name = str(i)
                if str(i).lower().startswith(s):
                    self.all_list_choices.append(self.generate_new(name, col))
            col.update()

        page: flet.Page = self.main_class.page

        cont = flet.Container(bgcolor=flet.colors.BLACK, border_radius=12)
        loading_text = flet.Text("Loading..", color=flet.colors.WHITE)
        cont.content = flet.Row([loading_text], alignment=flet.MainAxisAlignment.CENTER, height=100)

        page.dialog = flet.AlertDialog()
        page.dialog.content = cont
        page.dialog.open = True
        page.update()

        col = flet.ListView(width=450, height=500)
        col.controls.append(flet.Text("Choose one:", color="white", size=28, weight=flet.FontWeight.BOLD))

        search_field = flet.TextField(label="Search", on_change=on_search)
        col.controls.append(search_field)

        self.all_list_choices = []
        for icon in self.the_list:
            name = str(icon)
            if name.startswith("__"):
                pass
            else:
                self.all_list_choices.append(self.generate_new(name, col))

        cont.content = col
        page.update()

    def generate_new(self, value_name, col):
        def on_choose(e):
            self.self_ui.text = f"{self.title}: f{value_name}"
            self.__selected = value_name
            self.page.dialog.open = False
            self.page.update()

        r = flet.Row(
            [
                flet.Icon(value_name, size=22, color=flet.colors.WHITE),
                flet.Text(f"{value_name}", color=flet.colors.WHITE)
            ]
        )
        cc = flet.Container(content=r, on_click=on_choose)
        col.controls.append(cc)
        return cc

    @property
    def value(self):
        return self.__selected



import flet
from ...tools.color_picker import ColorPicker
from ...tools.list_picker import ListPopup


class Label(object):
    def __init__(self, main_class, parent, *args, **kwargs) -> None:
        self.parent = parent
        self.main_class = main_class
        self.self_object = flet.Row()
        self.title_object = flet.Text()
        self.icon_object = flet.Icon()

        self.self_object.controls.append(self.icon_object)
        self.self_object.controls.append(self.title_object)

        # all args
        all_icons = []
        for icon in flet.icons.__dict__:
            name = str(icon)
            if name.startswith("__"):
                pass
            else:
                all_icons.append(name)

        self.args = {
            "title": {"type": str, "default_value": "Just me and coffee"},
            "icon": {"type": ListPopup, "options": all_icons, "default_value": "EMOJI_FOOD_BEVERAGE_ROUNDED"},
            "title_color": {"type": ColorPicker, "default_value": "white"},
            "icon_color": {"type": ColorPicker, "default_value": "white"},
            "title_size": {"type": int, "default_value": 18},
            "icon_size": {"type": int, "default_value": 18},
            "spacing": {"type": int, "default_value": 15},
            "title_bold": {"type": int, "default_value": True},
            "alignment": {"type": list, "options": ["left", "center", "right"], "default_value": "center"}
        }

        # Template dict
        # This is where the widget data will be stored.
        self.template = {
            "widget_class_name": "Label",
            "properties": {}
        }
        for p in self.args:
            self.template["properties"][p] = self.args[p]["default_value"]

        self.update()

    def update(self, new_props: dict = None):
        r = self.self_object
        t = self.title_object
        I = self.icon_object
        props = self.template["properties"]

        if new_props is not None:
            for i in new_props:
                self.template["properties"][i] = new_props[i]

        r.spacing = props["spacing"]

        t.value = props["title"]
        t.color = props["title_color"]
        t.size = props["title_size"]

        if props["title_bold"]:
            t.weight = "bold"
        else:
            t.weight = "normal"

        I.name = props["icon"]
        I.color = props["icon_color"]
        I.size = props["icon_size"]

        if self.self_object.page is not None:
            self.self_object.update()

    def return_widget(self):
        props = self.template["properties"]
        if props["alignment"] == "left":
            return flet.Row([flet.Text("    "), self.self_object])
        elif props["alignment"] == "center":
            return flet.Row([self.self_object], alignment=flet.MainAxisAlignment.CENTER)
        else:
            return flet.Row([self.self_object, flet.Text("    ")], alignment=flet.alignment.center_right)



import flet
from ...tools.color_picker import ColorPicker


class Title(object):
    def __init__(self, main_class, parent, *args, **kwargs) -> None:
        self.parent = parent
        self.main_class = main_class
        self.self_object = flet.Text("")

        # all args
        self.args = {
            "title": {"type": str, "default_value": "myTitle"},
            "title_color": {"type": ColorPicker, "default_value": "white"},
            "size": {"type": int, "default_value": 23},
            "width": {"type": int, "default_value": 350},
            "italic": {"type": bool, "default_value": False},
            "bold": {"type": bool, "default_value": True},
            "hide": {"type": bool, "default_value": False},
            "expand": {"type": bool, "default_value": False},
            "alignment": {"type": list, "options": ["left", "center", "right"], "default_value": "center"},
            "text_align": {"type": list, "options": ["left", "center", "right"], "default_value": "left"}
        }

        # Template dict
        # This is where the widget data will be stored.
        self.template = {
            "widget_class_name": "Title",
            "properties": {}
        }
        for p in self.args:
            self.template["properties"][p] = self.args[p]["default_value"]

        self.update()

    def update(self, new_props: dict = None):
        t = self.self_object
        props = self.template["properties"]

        if new_props is not None:
            for i in new_props:
                self.template["properties"][i] = new_props[i]

        t.value = props["title"]
        t.size = props["size"]
        t.color = props["title_color"]
        t.width = props["width"]
        t.expand = props["expand"]
        t.italic = props["italic"]
        t.visible = props["hide"] == False
        if str(props["text_align"]).lower() == "left":
            t.text_align = flet.TextAlign.LEFT
        elif str(props["text_align"]).lower() == "center":
            t.text_align = flet.TextAlign.CENTER
        elif str(props["text_align"]).lower() == "right":
            t.text_align = flet.TextAlign.RIGHT

        if props["bold"]:
            t.weight = "bold"
        else:
            t.weight = "normal"

        if self.self_object.page is not None:
            self.self_object.update()

    def return_widget(self):
        props = self.template["properties"]
        if props["alignment"] == "left":
            return flet.Row([flet.Text("    "), self.self_object])
        elif props["alignment"] == "center":
            return flet.Row([self.self_object], alignment=flet.MainAxisAlignment.CENTER)
        else:
            return flet.Row([self.self_object, flet.Text("    ")], alignment=flet.alignment.center_right)



import flet
from ...tools.color_picker import ColorPicker


class Navigator(object):
    def __init__(self, main_class, parent, *args, **kwargs) -> None:
        self.parent = parent
        self.main_class = main_class
        self.self_object = flet.ElevatedButton(on_click=self.on_button_click)

        # all args
        all_pages = []
        for i in self.main_class.dict_content["pages"]:
            all_pages.append(i)
        self.args = {
            "text": {"type": str, "default_value": "click me"},
            "page name": {"type": list, "options": all_pages, "default_value": "main"},
            "width": {"type": int, "default_value": 150},
            "height": {"type": int, "default_value": 45},
            "alignment": {"type": list, "options": ["left", "center", "right"], "default_value": "center"},
            "bgcolor": {"type": ColorPicker, "default_value": "black"},
            "text_color": {"type": ColorPicker, "default_value": "blue"}
        }

        # Template dict
        # This is where the widget data will be stored.
        self.template = {
            "widget_class_name": "Navigator",
            "properties": {}
        }
        for p in self.args:
            self.template["properties"][p] = self.args[p]["default_value"]

        self.update()

    def update(self, new_props: dict = None):
        t = self.self_object
        props = self.template["properties"]

        if new_props is not None:
            for i in new_props:
                self.template["properties"][i] = new_props[i]

        t.text = props["text"]
        t.color = props["text_color"]
        t.bgcolor = props["bgcolor"]
        t.width = props["width"]
        t.height = props["height"]

        if self.self_object.page is not None:
            self.self_object.update()

    def on_button_click(self, event):
        if self.main_class.development_mode:
            return
        else:
            props = self.template["properties"]
            self.main_class.storyboard_class.navigate_to_page(props["page name"])

    def return_widget(self):
        props = self.template["properties"]
        if self.main_class.development_mode:
            self.self_object.disabled = True
        if props["alignment"] == "left":
            return flet.Row([flet.Text("    "), self.self_object])
        elif props["alignment"] == "center":
            return flet.Row([self.self_object], alignment=flet.MainAxisAlignment.CENTER)
        else:
            return flet.Row([self.self_object, flet.Text("    ")], alignment=flet.alignment.center_right)



from .engines.viewer_engine import viewerEngine
from .tools.storyboard_class import StoryBoard
from .widgets.All import all_widgets
import json
import flet


class LoadStoryBoard:
    def __init__(self, target_function, storyboard_file_path: str):
        # set up the important props
        self.current_page_name = "main"
        self.file_path = storyboard_file_path
        self.development_mode = False
        self.all_widgets = all_widgets
        self.dict_content = json.loads(open(self.file_path, encoding="utf-8").read())
        self.target_function = target_function

        # copy of things
        self.viewerEngine = viewerEngine
        # Run the app.
        flet.app(target=self.run)

    def run(self, page: flet.Page):
        self.storyboard_class = StoryBoard(page=page, main_class=self)
        page.vertical_alignment = page.vertical_alignment = flet.MainAxisAlignment.CENTER
        self.page = page
        ve = viewerEngine(self, self.dict_content, self.current_page_name, page, page, self.development_mode)
        self.target_function(self.storyboard_class)
        page.update()

    def setup_storyboard_class(self):
        pass



from .widgets.title import Title
from .widgets.open_url import Open_Url
from .widgets.button import Button
from .widgets.label import Label
from .widgets.markdown import Markdown
from .widgets.paragraph import Paragraph
from .widgets.row import Row
from .widgets.image import Image
from .widgets.padding import Padding
from .widgets.textfield import TextField
from .widgets.navigator import Navigator

all_widgets = {
    "Title": {"icon": "TEXT_FIELDS_ROUNDED", "class": Title},
    "Open Url": {"icon": "INSERT_LINK_SHARP", "class": Open_Url},
    "Button": {"icon": "SMART_BUTTON_ROUNDED", "class": Button},
    "Label": {"icon": "MEDICAL_INFORMATION_OUTLINED", "class": Label},
    "Markdown": {"icon": "TEXT_SNIPPET_ROUNDED", "class": Markdown},
    "Paragraph": {"icon": "SHORT_TEXT_ROUNDED", "class": Paragraph},
    "Row": {"icon": "VIEW_COLUMN_OUTLINED", "class": Row},
    "Image": {"icon": "IMAGE_OUTLINED", "class": Image},
    "Padding": {"icon": "SPACE_BAR_ROUNDED", "class": Padding},
    "TextField": {"icon": "INPUT_ROUNDED", "class": TextField},
    "Navigator": {"icon": "NAVIGATE_NEXT", "class": Navigator}
}



#! This is the editing page of a storyboard.
from flet import Page, Row, Text
from ..widgets.All import all_widgets
import flet
import os
import json
import time
import requests


#* local imports
from ..sections.left_section import leftSection
from ..sections.preview_section import PreviewSection
from ..sections.edit_section import editSection
from ..engines.edit_widget_engine import EditWidgetsEngine
from ..engines.edit_subwidgets_engine import EditSubWidgetsEngine
from ..engines.suggesting_engine import SuggestingEngine
from ..widgets.All import all_widgets
from ..pages.settings import SettingsPage


class mainPage:
    def __init__(self, file_path):
        if not os.path.isfile(file_path):
            raise FileExistsError(f"There is no Flet StoryBoard on path '{file_path}' .")
        

        self.current_page_name = "main"
        self.file_path = file_path
        self.development_mode = True
        self.dict_content = json.loads(open(file_path, encoding="utf-8").read())

        #? Copy of classes
        self.all_widgets = all_widgets
        self._editWidgetsEngine = EditWidgetsEngine
        self._editSubWidgetsEngine = EditSubWidgetsEngine

        #? Run the app
        flet.app(target=self.app)

    def app (self, page:Page):
        page.title = f"Flet StoryBoard - {self.file_path}"
        page.spacing = 0
        page.bgcolor = "black"
        page.vertical_alignment = flet.MainAxisAlignment.CENTER
        page.window_width = 850
        page.window_height = 650
        page.window_min_width = 850
        page.window_min_height = 640
        self.page = page
        page.appbar = self.generate_app_bar()
        page.window_center()
        page.update()

        # The main row
        self.main_row = Row(scroll=False)
        page.add(self.main_row)

        #? append the sections
        self.left_section = leftSection(page, self, self.main_row)
        self.preview_section = PreviewSection(page, self, self.main_row)
        self.edit_section = editSection(page, self, self.main_row)

        #? Set finals of page.
        page.on_resize = self.on_page_resize
        page.on_keyboard_event = self.manage_keyboard_commands


        #? Setup the storyboard suggestions engine.
        self.suggesting_engine = SuggestingEngine(self)
        time.sleep(0.5)
        self.suggesting_engine.push_new_suggestion()

    def on_page_resize (self, *event):
        page = self.page
        self.left_section.main_container.width = self.page.width / 4 - 30
        self.left_section.main_container.height = self.page.height - 70
        self.preview_section.main_view.width = page.width-(page.width/4)*2
        self.preview_section.main_view.height = page.height-150
        self.edit_section.main_column.width = self.page.width / 4 - 10
        self.edit_section.main_column.height = page.height - 70

        self.page.update()
    
    def pages_browser (self):
        def create_a_page (e):
            self.create_new_page(str(e.control.value))
            self.page.appbar = self.generate_app_bar()
            self.page.update()
        def open_a_page (e):
            page_name = str(e.control.content.controls[0].value)
            if page_name in self.dict_content["pages"]:
                self.current_page_name = page_name
            self.preview_section.update_preview(self.current_page_name)
            self.last_checked_page_button.bgcolor = flet.colors.GREY_700
            self.last_checked_page_button = e.control
            e.control.bgcolor = "blue"
            time.sleep(0.3)
            self.suggesting_engine.push_new_suggestion()
            self.page.update()
        def ask_for_new_page_name (e):
            mr.scroll = False
            tf = flet.TextField(label="Page name", on_submit=create_a_page, height=40, color="white")
            mr.controls = [tf]
            mr.update()
            tf.focus()
        mr = flet.Row([], width=250, scroll=True)
        mr.controls.append(flet.TextButton(content=flet.Text("✨", size=18), on_click=self.edit_page_suggestion_state, width=35))
        new_page_button = flet.Container(flet.Row([flet.Text("+ Page", size=12, color="black")], alignment="center"), on_click=ask_for_new_page_name, 
        bgcolor="white", width=60, height=30, border_radius=12)
        mr.controls.append(new_page_button)

        for p in self.dict_content["pages"]:
            c = flet.Container(flet.Row([
                flet.Text(p, color="white", size=12)
            ], alignment="center"), bgcolor=flet.colors.GREY_700, width=60, height=30, border_radius=12, on_click=open_a_page)
            mr.controls.append(c)
            if str(p) == str(self.current_page_name):
                c.bgcolor = "blue"
                self.last_checked_page_button = c

        return mr
    
    def generate_app_bar (self):
        a = flet.AppBar(
            bgcolor=self.page.bgcolor,
            leading=flet.Row([
                Text("    "),
                flet.Icon(flet.icons.DASHBOARD_ROUNDED, size=18, color="white"),
                Text("Flet StoryBoard", color="white", weight="bold", size=15)
            ], spacing=15
            ),
            actions=[
                Row([
                    flet.TextButton(content=flet.Text("Settings", size=12, color="white"), 
                    on_click=self.open_settings_page),
                    flet.ElevatedButton("Save", bgcolor="white", color="black", width=100, height=35, 
                    on_click=self.save_all, tooltip="click to save | also control + S"),
                    Text("    ")
                ], alignment=15)
            ],
            title=self.pages_browser(),
            center_title=True
        )

        return a
    

    def manage_keyboard_commands (self, event):
        key = event.key # It would be a key like: 'A' or 'Enter'.
        shift = event.shift # It would be a bool value of where the shift key is clicked or not.
        ctrl = event.ctrl # It would be a bool value of where the control key is clicked or not.
        alt = event.alt # It would be a bool value of where the option key is clicked or not.
        meta = event.meta # It would be a bool value of where the command key is clicked or not.

        if str(key).lower() == "s" and ctrl:
            #? Clicked to save file
            self.save_all()
        elif str(key).lower() == "s" and meta:
            #? Clicked to save file
            self.save_all()
    

    def save_all (self, *args):
        new_content = json.dumps(self.dict_content)
        file = open(self.file_path, "w+", encoding="utf-8")
        file.write(new_content)

        page = self.page
        page.snack_bar = flet.SnackBar(
            content=flet.Text(f"This Flet StoryBoard '{self.file_path}' is saved!"),
            action="Alright!"
        )
        page.snack_bar.open = True
        page.update()
    

    def add_new_widget (self, widget_name, page_name=None):
        widget_number = len(self.dict_content["pages"][self.current_page_name]["widgets"])-1
        widget_class = all_widgets[widget_name]["class"]
        widget_class = widget_class(self, self.preview_section.main_view_column, widget_number=widget_number)
        
        self.dict_content["pages"][self.current_page_name]["widgets"].append(widget_class.template)
        
        self.preview_section.update_preview(self.current_page_name)
        self.page.update()

        self.edit_a_widget(len(self.dict_content["pages"][self.current_page_name]["widgets"])-1)
        
        time.sleep(0.5)
        self.suggesting_engine.push_new_suggestion()

        return widget_class
    
    def edit_a_widget (self, widget_number_on_content):
        self.edit_section.edit_widget_using_engine(widget_number_on_content)
        self.page.update()
    

    def create_new_page (self, page_name):
        page_dict = {
            "settings" : {
                "bgcolor" : "black",
                "suggestions_rules" : "none"
            },
                "widgets" : [
                            
            ]
        }
        self.dict_content["pages"][page_name] = page_dict
        self.current_page_name = str(page_name)
        self.page.appbar = self.generate_app_bar()
        self.preview_section.update_preview(self.current_page_name)
    

    def edit_page_suggestion_state (self, *a):
        def go_back (e):
            self.left_section.show_new_content(last_content)
            self.page.update()
        
        def apply_and_start (name_of_sug):
            self.dict_content["pages"][self.current_page_name]["settings"]["suggestions_rules"] = str(name_of_sug)
            go_back("")
            self.suggesting_engine.push_new_suggestion()

        def make_a_rule_button_option (name):
            def on_choose (e):
                apply_and_start(name)
            b = flet.ElevatedButton(f"{name}", on_click=on_choose, bgcolor="white", color="blacK", width=150, height=45)
            return flet.Row([b], alignment="center")

        last_content = self.left_section.main_container
        c = flet.Column(width=last_content.width, height=last_content.height)

        back_btn = flet.TextButton("< Back", on_click=go_back)
        c.controls.append(back_btn)

        title = flet.Text("Chose the suggestion rules for this page.", color="white", size=17, weight="bold")
        c.controls.append(title)

        all_rules = requests.get("https://raw.githubusercontent.com/SKbarbon/Flet_StoryBoard/main/rules/all.json").text
        all_rules = json.loads(all_rules)
        
        for rule_option in all_rules["all_rules"]:
            c.controls.append(make_a_rule_button_option(rule_option))

        c.controls.append(flet.Row([flet.Text("more will come soon", color="white", size=13)], alignment="center"))
        c.scroll = True
        self.left_section.show_new_content(c)
    
    def open_settings_page (self, *e):
        SettingsPage(self.page, self)


from bs4 import BeautifulSoup
import requests
import urllib3
from requests.packages.urllib3.exceptions import InsecureRequestWarning

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def get_site_favicon(url):
    response = requests.get(url, verify=False)
    soup = BeautifulSoup(response.text, 'html.parser')
    favicon_link = soup.find('link', rel='icon') or soup.find('link', rel='shortcut icon') or soup.find('link', rel='apple-touch-icon')
    try:
        favicon_url = favicon_link['href']
        if str(favicon_url).startswith("http"):
            return f"{favicon_url}"
        else:
            return f"{url}/{favicon_url}"
    except:
        return None



from flet import Page
import flet
import time

from .Settings.pages import page_settings_page

class SettingsPage:
    def __init__(self, page:Page, main_class) -> None:
        self.page : Page = page
        self.main_class = main_class
        #? Create a copy of things.
        self.__last_keyboard_manager = main_class.manage_keyboard_commands
        self.__last_controls = list(page.controls)
        self.__lat_appbar = page.appbar
        self.__last_on_resize_function = main_class.on_page_resize
        
        #? For keyboard events
        self.current_selected_page = None
        self.page_opend = True
        #? Hide all page controls
        self.hide_all()

        #? ReSet all page controls
        page.on_keyboard_event = self.keyboard_keys_manager
        page.on_resize = self.on_page_resize
        page.controls.clear()
        page.appbar = None
        page.bgcolor = "black"
        page.update()

        main_row = flet.Row(alignment="center")
        page.add(main_row)

        #? pages browseing section
        pages_browser_container = flet.Container(border_radius=12)
        self.pages_broswer = flet.Column(width=page.width/3, height=page.height-50, scroll=True)
        pages_browser_container.content = self.pages_broswer
        main_row.controls.append(pages_browser_container)
        page.update()

        #? viewing selected page section
        page_viewing_section_container = flet.Container(bgcolor="#1C1C1E", border_radius=12)
        self.page_viewing_section = flet.Column(width=page.width-(page.width/3)-50, height=page.height-50, scroll=True)
        page_viewing_section_container.content = self.page_viewing_section
        main_row.controls.append(page_viewing_section_container)
        page.update()

        #? The title and back_btn
        title = flet.Text(" Settings", size=28, weight="bold", color="white")
        self.pages_broswer.controls.append(flet.Row([
            flet.IconButton("CLOSE_ROUNDED", on_click=self.go_back, icon_color="white", width=30), title
        ], spacing=0))

        self.pages_broswer.controls.append(flet.Text(""))

        #? show all settings pages
        for p in self.get_all_settings_pages():
            self.pages_broswer.controls.append(page_navigator_frame_button(p, self.get_all_settings_pages()[p]["icon"], self.get_all_settings_pages()[p]["function"],self))
        
        page.update()

    def open_new_settings_page(self, control):
        """Open a new sub-page on settings page"""
        pass

    def route_to_page (self):
        """Use flet Route to a page"""
    
    def hide_all (self):
        page : Page = self.page
        for i in range(10):
            for c in page.controls:
                c.opacity = c.opacity - 0.1
                page.appbar.opacity = page.appbar.opacity - 0.1
                c.update()
            for ac in page.appbar.actions:
                ac.opacity = ac.opacity - 0.1
                ac.update()
            page.appbar.leading.opacity = page.appbar.leading.opacity - 0.1
            page.appbar.leading.update()
            page.appbar.title.opacity = page.appbar.title.opacity - 0.1
            page.appbar.title.update()
            time.sleep(0.01)

    def go_back (self, *e):
        self.page_opend = False
        page : Page = self.page
        page.on_keyboard_event = self.__last_keyboard_manager
        page.controls.clear()
        page.bgcolor = "black"
        page.appbar = self.__lat_appbar

        for c in self.__last_controls:
            c.opacity = 1.0
            page.controls.append(c)
        
        for ac in page.appbar.actions:
            ac.opacity = 1.0

        page.appbar.leading.opacity = 1.0
        page.appbar.title.opacity = 1.0
        
        page.on_resize = self.__last_on_resize_function
        page.update()

    def on_page_resize (self, e):
        page = self.page
        self.pages_broswer.width = page.width/3 - 50
        self.pages_broswer.height = self.page.height - 50
        self.page_viewing_section.width = page.width-(page.width/3-50)-50
        self.page_viewing_section.height = page.height-50
        self.page.update()
    
    def get_all_settings_pages (self):
        return {
            "Pages" : {"icon":"PREVIEW_ROUNDED", "function":page_settings_page, "name":"Pages"},
            "Editor" : {"icon":"SETTINGS_ROUNDED", "function":editor_page, "name":"Editor"}
        }
    
    def keyboard_keys_manager (self, e):
        if self.page_opend == False: return
        pass


#? Generate page button
def page_navigator_frame_button (name:str, icon, function, settings_class:SettingsPage, as_a_click=False):
    def on_hov (e):
        if str(e.data) == "true":
            container.bgcolor = flet.colors.WHITE60
        else:
            container.bgcolor = None
        container.update()
        if e.control == settings_class.current_selected_page:
            e.control.bgcolor = "blue"
            e.control.update()
    
    def on_click (e):
        if settings_class.current_selected_page == None:
            settings_class.current_selected_page = container
            container.bgcolor = "blue"
            if container.page == None:
                settings_class.page.update()
            else:
                container.update()
        else:
            settings_class.current_selected_page.bgcolor = None
            if settings_class.current_selected_page.page != None:
                settings_class.current_selected_page.update()
            settings_class.current_selected_page = container
            container.bgcolor = "blue"
            if container.page == None:
                settings_class.page.update()
            else:
                container.update()
            
        function(settings_class)
        
        

    container = flet.Container(width=200, height=40, border_radius=12, expand=True, 
    on_hover=on_hov, on_click=on_click)
    r = flet.Row([flet.Text("  ")], alignment="center", width=200)
    container.content = r

    ICON = flet.Icon(icon, size=20, color="white")
    r.controls.append(ICON)

    NAME = flet.Text(name, size=17, color="white", expand=True, text_align="left")
    r.controls.append(NAME)

    if as_a_click:
        container.bgcolor = "blue"
        on_click(container)

    return flet.Row([container], alignment="center")


#? Pages


def editor_page (settings_class:SettingsPage):
    v = settings_class.page_viewing_section
    v.clean()
    v.controls.append(flet.Text("soon", color="white"))
    v.update()


import json


def Create_StoryBoard(file_name, template="default", storyboard_suggestions=False, main_page_suggestions_rules="none"):
    if template == "default":
        storyboard_default_template = {
            "storyboard_settings": {
                "template": "default",
                "storyboard_suggestions": storyboard_suggestions,
                "allow_scroll": False
            },
            "pages": {
                "main": {
                    "settings": {
                        "bgcolor": "black",
                        "suggestions_rules": f"{main_page_suggestions_rules}"
                    },
                    "widgets": [

                    ]
                }
            }
        }

        file = open(f"{file_name}.fletsb", "w+", encoding="utf-8")
        file.write(json.dumps(storyboard_default_template))
        return True
    else:
        print("External templates are nor supported yet.")



import flet

from ..widgets.All import all_widgets
from ..tools.color_picker import ColorPicker
from ..tools.list_picker import ListPopup


class EditWidgetsEngine:
    def __init__(self, main_class, section_view: flet.Column, widget_number) -> None:
        self.main_class = main_class
        self.section_view = section_view
        self.widget_number = widget_number

        self.current_page_name = self.main_class.current_page_name
        self.all_fields = {}

        self.show_edit_tools()

    def show_edit_tools(self):
        self.section_view.clean()
        content = self.main_class.dict_content
        class_name = content["pages"][self.current_page_name]["widgets"][self.widget_number]["widget_class_name"]
        widget_content = content["pages"][self.current_page_name]["widgets"][self.widget_number]

        if class_name not in all_widgets:
            # if widget not found.
            self.section_view.controls.append(
                flet.Row(
                    [flet.Text(f"There is no supported\nwidget named '{class_name}'.", color=flet.colors.WHITE)]
                )
            )
            return

        the_class = all_widgets[class_name]["class"]
        the_class = the_class(
            self.main_class,
            self.main_class.preview_section.main_view,
            widget_number=self.widget_number
        )
        default_args = the_class.args

        title = flet.Text(f"Edit {class_name}", size=25, weight=flet.FontWeight.BOLD, color=flet.colors.WHITE)
        self.section_view.controls.append(title)

        for t in widget_content["properties"]:
            prop_name = t
            prop_value = widget_content["properties"][t]
            if prop_name not in default_args:
                print("Error while load widget properties in edit engine.")
                self.section_view.clean()
                return
            prop_type = default_args[prop_name]["type"]

            # These statements below are for adding the fields for edit a widget.
            if isinstance(prop_type(), str):
                tf = flet.TextField(width=160, bgcolor=flet.colors.WHITE, color=flet.colors.BLACK, label=prop_name)
                tf.value = prop_value
                self.section_view.controls.append(flet.Row([tf], alignment=flet.MainAxisAlignment.CENTER))
                self.all_fields[prop_name] = tf
                if "multi_line" in default_args[prop_name]:
                    tf.multiline = True

            elif isinstance(prop_type(), int):
                slid = flet.Slider(min=0, max=500, divisions=500, label="{value}", width=160)
                slid.value = int(prop_value)
                self.section_view.controls.append(
                    flet.Row(
                        [
                            flet.Text(
                                f"{prop_name}:",
                                color=flet.colors.WHITE,
                                width=150
                            )
                        ], alignment=flet.MainAxisAlignment.CENTER
                    )
                )
                self.section_view.controls.append(
                    flet.Row(
                        [slid],
                        alignment=flet.MainAxisAlignment.CENTER
                    )
                )
                self.all_fields[prop_name] = slid

            elif isinstance(prop_type(), bool):
                tog = flet.Switch()
                tog.value = prop_value
                self.section_view.controls.append(
                    flet.Row([flet.Text(f"{prop_name}", color=flet.colors.WHITE, size=13), tog],
                             alignment=flet.MainAxisAlignment.CENTER, spacing=25))
                self.all_fields[prop_name] = tog

            elif type(prop_type()) == type(ColorPicker()):
                self.section_view.controls.append(flet.Text(""))
                colp = ColorPicker(self.section_view, selected_color=prop_value, add_it=False, title_name=prop_name,
                                   drop_width=120)
                self.section_view.controls.append(flet.Row([colp.v], alignment=flet.MainAxisAlignment.CENTER))
                self.all_fields[prop_name] = colp

            elif isinstance(prop_type(), list):
                dr = flet.Dropdown(width=160, label=prop_name)
                for i in default_args[prop_name]["options"]:
                    dr.options.append(flet.dropdown.Option(f"{i}"))
                dr.value = prop_value
                self.section_view.controls.append(flet.Row([dr], alignment=flet.MainAxisAlignment.CENTER))
                self.all_fields[prop_name] = dr

            elif type(prop_type()) == type(ListPopup()):
                lp = ListPopup(default_args[prop_name]["options"], self.main_class, prop_value, prop_name)
                self.section_view.controls.append(flet.Row([lp.self_ui], alignment=flet.MainAxisAlignment.CENTER))
                self.all_fields[prop_name] = lp

        # This button below for widgets that support sub-widgets.
        if hasattr(the_class, "support_sub_widgets"):
            add_sub_widget_button = flet.TextButton("Add sub-widget", on_click=the_class.widgets_to_add_in)
            self.section_view.controls.append(
                flet.Row(
                    [add_sub_widget_button],
                    alignment=flet.MainAxisAlignment.CENTER
                )
            )

        # This slider bellow is for rearrange the widget on the page.
        rearrange_slider = flet.Slider(
            min=0,
            max=len(content["pages"][self.current_page_name]["widgets"]),
            value=self.widget_number,
            label="{value}",
            width=160,
            divisions=len(content["pages"][self.current_page_name]["widgets"])
        )
        self.section_view.controls.append(
            flet.Row(
                [flet.Text(f"reArrange:", color=flet.colors.WHITE, width=150)],
                alignment=flet.MainAxisAlignment.CENTER
            )
        )
        self.section_view.controls.append(flet.Row([rearrange_slider], alignment=flet.MainAxisAlignment.CENTER))
        self.rearrange_slider = rearrange_slider

        self.section_view.controls.append(flet.Text("\n"))
        # Down bellow is for done_button and delete_btn.
        done_button = flet.ElevatedButton(
            "Done",
            bgcolor=flet.colors.WHITE,
            color=flet.colors.BLACK,
            width=150,
            height=40,
            on_click=self.done_edit
        )
        self.section_view.controls.append(
            flet.Row(
                [done_button],
                alignment=flet.MainAxisAlignment.CENTER
            )
        )

        delete_button = flet.TextButton(
            content=flet.Text("delete", color=flet.colors.RED, size=13),
            on_click=self.delete_widget
        )
        self.section_view.controls.append(
            flet.Row(
                [delete_button],
                alignment=flet.MainAxisAlignment.CENTER
            )
        )

    def done_edit(self, *args):
        new_widget_properties_dict = {}
        for P in self.all_fields:
            new_widget_properties_dict[P] = self.all_fields[P].value

        if int(self.widget_number) == int(self.rearrange_slider.value):
            # If the widget arrange is the same.
            self.main_class.dict_content["pages"][self.current_page_name]["widgets"][self.widget_number][
                "properties"].update(new_widget_properties_dict)
        else:
            # If the widget arrange is NOT the same.
            if int(self.rearrange_slider.value) >= int(
                    len(self.main_class.dict_content["pages"][self.main_class.current_page_name]["widgets"])):
                self.main_class.dict_content["pages"][self.current_page_name]["widgets"][self.widget_number][
                    "properties"].update(new_widget_properties_dict)
                copy_of_content = self.main_class.dict_content["pages"][self.current_page_name]["widgets"][
                    self.widget_number]
                del self.main_class.dict_content["pages"][self.current_page_name]["widgets"][self.widget_number]

                self.main_class.dict_content["pages"][self.current_page_name]["widgets"].append(copy_of_content)

                self.widget_number = int(
                    len(self.main_class.dict_content["pages"][self.current_page_name]["widgets"]) - 1)
            else:
                self.main_class.dict_content["pages"][self.current_page_name]["widgets"][self.widget_number][
                    "properties"].update(new_widget_properties_dict)
                copy_of_content = self.main_class.dict_content["pages"][self.current_page_name]["widgets"][
                    self.widget_number]

                self.main_class.dict_content["pages"][self.current_page_name]["widgets"][
                    int(self.rearrange_slider.value)] = copy_of_content
                del self.main_class.dict_content["pages"][self.current_page_name]["widgets"][self.widget_number]

                self.widget_number = int(self.rearrange_slider.value)

        # Update the viewer engine to see the edit changes.
        self.main_class.preview_section.update_preview(self.main_class.current_page_name)
        self.main_class.page.update()

        # ReOpen the editor again
        EditWidgetsEngine(
            main_class=self.main_class,
            section_view=self.section_view,
            widget_number=self.widget_number
        )
        self.main_class.page.update()

    def delete_widget(self, *args):
        del self.main_class.dict_content["pages"][self.current_page_name]["widgets"][self.widget_number]
        self.section_view.clean()

        # Update the viewer engine to see the edit changes.
        self.main_class.preview_section.update_preview(self.main_class.current_page_name)
        self.main_class.page.update()



import json
import flet
import random
import os
import requests


class SuggestingEngine:
    def __init__(self, main_class):
        self.main_class = main_class

        self.main_col = flet.Column(
            width=self.main_class.left_section.main_container.width,
            height=self.main_class.left_section.main_container.height,
            # issue: True is not a valid value for the scroll mode. Use flet.SCROLL_MODE
            scroll=True
        )

    def push_new_suggestion(self):
        settings = self.main_class.dict_content["storyboard_settings"]
        applied = settings["storyboard_suggestions"]

        # if `storyboard suggestions` is off.
        if not applied:
            return

        suggestions_rules_name = self.main_class.dict_content["pages"][self.main_class.current_page_name]["settings"][
            "suggestions_rules"]

        if suggestions_rules_name == "none":
            return

        # Get the rules file
        if not os.path.isdir("rules"):
            os.mkdir("rules")

        if not os.path.isfile(f"rules/{suggestions_rules_name}.json"):
            all_rules = requests.get("https://raw.githubusercontent.com/SKbarbon/Flet_StoryBoard/main/rules/all.json").text
            all_rules = json.loads(all_rules)

            if suggestions_rules_name not in all_rules["all_rules"]:
                return
            else:
                rf = requests.get(all_rules["all_rules"][suggestions_rules_name]).text
                open(f"rules/{suggestions_rules_name}.json", "w+", encoding="utf-8").write(rf)

        rules_file = json.loads(open(f"rules/{suggestions_rules_name}.json", encoding="utf-8").read())

        # step1: Get current widgets.
        current_widgets = list(self.main_class.dict_content["pages"][self.main_class.current_page_name]["widgets"])

        # step2: Count classes on widgets.
        classes_in_widgets = {}
        for c in current_widgets:
            if c["widget_class_name"] not in classes_in_widgets:
                classes_in_widgets[c["widget_class_name"]] = 1
            else:
                classes_in_widgets[c["widget_class_name"]] = classes_in_widgets[c["widget_class_name"]] + 1

        # step3: Search for the correct case on the rules file.
        for rule in rules_file["rules"]:
            if rule["case"] == classes_in_widgets:
                self.change_left_section_into_suggestion(rule["sugs"])
                break

    def change_left_section_into_suggestion(self, suggestions):
        def go_back(e):
            if last_on_section == self.main_col:
                self.main_class.left_section.show_all_widgets()
            else:
                self.main_class.left_section.show_new_content(last_on_section)

        main_col = self.main_col
        main_col.controls.clear()

        back_btn = flet.TextButton("< Back", on_click=go_back)
        main_col.controls.append(back_btn)

        title = flet.Text("Suggestions ✨", weight=flet.FontWeight.BOLD, size=22, color=flet.colors.WHITE)
        main_col.controls.append(title)

        main_col.controls.append(
            flet.Row(
                [flet.Text("", color=flet.colors.WHITE, size=13)],
                alignment=flet.MainAxisAlignment.CENTER
            )
        )

        for sug in suggestions:
            sc = self.suggestion_card(sug, go_back)
            main_col.controls.append(
                flet.Row(
                    [sc],
                    alignment=flet.MainAxisAlignment.CENTER
                )
            )

        last_on_section = self.main_class.left_section.main_container
        self.main_class.left_section.show_new_content(self.main_col)

    def suggestion_card(self, suggestion_dict, go_back_function):
        def add_widget(e):
            go_back_function(e)
            w = self.main_class.add_new_widget(suggestion_dict["class"])
            w.update(suggestion_dict["props"])
            self.main_class.preview_section.update_preview(self.main_class.current_page_name)
            self.main_class.edit_a_widget(len(self.main_class.dict_content["pages"][self.main_class.current_page_name]["widgets"]) - 1)
            self.main_class.page.update()

        colors = [flet.colors.WHITE, "#F5DEE5", flet.colors.YELLOW]

        card_container = flet.Container(
            bgcolor=random.choice(colors),
            width=170,
            height=225,
            border_radius=13
        )
        column = flet.Column()
        card_container.content = column

        class_name = suggestion_dict["class"]
        why_to_add = suggestion_dict["about"]

        title = flet.Text(
            value=f"\n   {class_name}",
            size=23,
            weight=flet.FontWeight.BOLD,
            color=flet.colors.BLACK,
            width=card_container.width
        )
        about = flet.Text(
            value=f"{why_to_add}",
            size=13,
            color=flet.colors.BLACK,
            width=card_container.width - 37,
            height=90
        )
        apply_button = flet.Container(
            flet.Row(
                [flet.Text("Add", color="white")],
                alignment=flet.MainAxisAlignment.CENTER
            ),
            bgcolor=flet.colors.BLACK,
            width=card_container.width - 15,
            height=40,
            border_radius=13,
            on_click=add_widget
        )

        column.controls.append(title)
        column.controls.append(flet.Row([about], alignment=flet.MainAxisAlignment.CENTER))
        column.controls.append(
            flet.Row(
                [apply_button],
                alignment=flet.MainAxisAlignment.CENTER
            )
        )

        # issue: column is not updated: column.update()

        return card_container



from flet import Column, Page, Row
import flet

from ..engines.edit_widget_engine import EditWidgetsEngine


class editSection:
    def __init__(self, page:Page, main_class, main_row:Row) -> None:
        self.page : Page = page
        self.main_class = main_class
        self.main_row : Row = main_row

        self.main_column = Column(width=page.width/4, height=page.height,scroll=True)

        main_row.controls.append(self.main_column)
    
    def edit_widget_using_engine (self, widget_number):
        ee = EditWidgetsEngine(main_class=self.main_class, section_view=self.main_column, widget_number=widget_number)


import flet
from ...tools.color_picker import ColorPicker


class TextField(object):
    def __init__(self, main_class, parent, *args, **kwargs) -> None:
        self.parent = parent
        self.main_class = main_class
        self.self_object = flet.TextField(
            on_focus=self.on_start_type,
            on_change=self.on_change_text,
            on_submit=self.on_end_type,
            selection_color=flet.colors.BLACK,
            cursor_color=flet.colors.BLACK,
            # issue: focused_color can't have value "color"
            focused_color="color"
        )

        # all args
        self.args = {
            "text": {"type": str, "default_value": ""},
            "label": {"type": str, "default_value": "text.."},
            "hint_text": {"type": str, "default_value": "text.."},
            "can_reveal_password": {"type": bool, "default_value": True},
            "password": {"type": bool, "default_value": False},
            "width": {"type": int, "default_value": 200},
            "height": {"type": int, "default_value": 65},
            "border_radius": {"type": int, "default_value": 12},
            "bgcolor": {"type": ColorPicker, "default_value": "white"},
            "color": {"type": ColorPicker, "default_value": "black"},
            "alignment": {"type": list, "options": ["left", "center", "right"], "default_value": "center"},
            "on start": {"type": str, "default_value": ""},
            "on change": {"type": str, "default_value": ""},
            "on end": {"type": str, "default_value": ""},
            "point name": {"type": str, "default_value": ""}
        }

        # Template dict
        # This is where the widget data will be stored.
        self.template = {
            "widget_class_name": "TextField",
            "properties": {}
        }
        for p in self.args:
            self.template["properties"][p] = self.args[p]["default_value"]

        self.update()

    def update(self, new_props: dict = None):
        tf = self.self_object
        props = self.template["properties"]

        if new_props is not None:
            for i in new_props:
                self.template["properties"][i] = new_props[i]

        tf.value = props["text"]
        tf.label = props["label"]
        tf.hint_text = props["hint_text"]
        tf.password = props["password"]
        tf.can_reveal_password = props["can_reveal_password"]
        tf.width = props["width"]
        tf.height = props["height"]
        tf.bgcolor = props["bgcolor"]
        tf.color = props["color"]
        tf.border_radius = props["border_radius"]

        if self.self_object.page is not None:
            self.self_object.update()

    def on_start_type(self, event):
        if self.main_class.development_mode:
            return
        else:
            props = self.template["properties"]
            if props["on start"] == "":
                return

            if props["on start"] in self.main_class.storyboard_class.functions:
                self.main_class.storyboard_class.functions[props["on start"]]()
            else:
                fn = props["on start"]
                print(f"Pass error: There is not function found called {fn}")

    def on_change_text(self, event):
        if self.main_class.development_mode:
            return
        else:
            props = self.template["properties"]
            if props["point name"] != "":
                self.main_class.storyboard_class.points[props["point name"]] = str(self.self_object.value)
            if props["on change"] == "":
                return

            if props["on change"] in self.main_class.storyboard_class.functions:
                self.main_class.storyboard_class.functions[props["on change"]](self.self_object.value)
            else:
                fn = props["on change"]
                print(f"Pass error: There is not function found called {fn}")

    def on_end_type(self, event):
        if self.main_class.development_mode:
            return
        else:
            props = self.template["properties"]
            if props["point name"] != "":
                self.main_class.storyboard_class.points[props["point name"]] = str(self.self_object.value)
            if props["on end"] == "": return

            if props["on end"] in self.main_class.storyboard_class.functions:
                self.main_class.storyboard_class.functions[props["on end"]](self.self_object.value)
            else:
                fn = props["on end"]
                print(f"Pass error: There is not function found called {fn}")

    def return_widget(self):
        if self.main_class.development_mode:
            self.self_object.disabled = True
        props = self.template["properties"]
        if props["alignment"] == "left":
            return flet.Row([flet.Text("    "), self.self_object])
        elif props["alignment"] == "center":
            return flet.Row([self.self_object], alignment=flet.MainAxisAlignment.CENTER)
        else:
            return flet.Row([self.self_object, flet.Text("    ")], alignment=flet.alignment.center_right)



import flet


def Widget_Browse_Frame(widget_name, widget_info_dict, on_click, page_name):
    def action(event):
        on_click(widget_name, page_name)

    def on_hover(event):
        event.control.bgcolor = "#474747" if event.data == "true" else "#5C5C5C"
        event.control.update()

    r = flet.Row(
        [
            flet.Text(" "),
            flet.Icon(widget_info_dict["icon"], size=18, color=flet.colors.WHITE),
            flet.Text(widget_name, size=13, color=flet.colors.WHITE)
        ],
        spacing=12
    )

    c = flet.Container(
        content=r,
        width=170,
        height=35,
        bgcolor="#5C5C5C",
        on_click=action,
        expand=True,
        on_hover=on_hover,
        border_radius=12
    )

    return flet.Row([c], alignment=flet.MainAxisAlignment.CENTER)



import flet
from ...tools.color_picker import ColorPicker


class Button(object):
    def __init__(self, main_class, parent, *args, **kwargs) -> None:
        self.parent = parent
        self.main_class = main_class
        self.self_object = flet.ElevatedButton(on_click=self.on_button_click)

        # all args
        self.args = {
            "text": {"type": str, "default_value": "click me"},
            "function name": {"type": str, "default_value": ""},
            "text_color": {"type": ColorPicker, "default_value": "black"},
            "bgcolor": {"type": ColorPicker, "default_value": "white"},
            "alignment": {"type": list, "options": ["left", "center", "right"], "default_value": "center"},
            "width": {"type": int, "default_value": 95},
            "height": {"type": int, "default_value": 40},
            "hide": {"type": bool, "default_value": False}
        }

        # Template dict
        # This is where the widget data will be stored.
        self.template = {
            "widget_class_name": "Button",
            "properties": {}
        }
        for p in self.args:
            self.template["properties"][p] = self.args[p]["default_value"]

        self.update()

    def update(self, new_props: dict = None):
        t = self.self_object
        props = self.template["properties"]

        if new_props is not None:
            for i in new_props:
                self.template["properties"][i] = new_props[i]

        t.text = props["text"]
        t.color = props["text_color"]
        t.bgcolor = props["bgcolor"]
        t.width = props["width"]
        t.height = props["height"]
        t.visible = props["hide"] == False

        if self.self_object.page is not None:
            self.self_object.update()

    def on_button_click(self, event):
        if self.main_class.development_mode:
            return
        else:
            props = self.template["properties"]
            if props["function name"] == "":
                return

            if props["function name"] in self.main_class.storyboard_class.functions:
                self.main_class.storyboard_class.functions[props["function name"]]()
            else:
                fn = props["function name"]
                print(f"Pass error: There is not function found called {fn}")

    def return_widget(self):
        props = self.template["properties"]
        if self.main_class.development_mode:
            self.self_object.disabled = True
        if props["alignment"] == "left":
            return flet.Row([flet.Text("    "), self.self_object])
        elif props["alignment"] == "center":
            return flet.Row([self.self_object], alignment=flet.MainAxisAlignment.CENTER)
        else:
            return flet.Row([self.self_object, flet.Text("    ")], alignment=flet.alignment.center_right)



from flet import Container, Row, Column
import flet


def none(*args):
    pass


class ColorPicker:

    def __init__(
            self,
            main_view=None,
            selected_color="white",
            on_choose_color=none,
            add_it=True,
            title_name="select color",
            drop_width=200,
            color_prev_width=50
    ) -> None:

        if main_view is None:
            return
        self.__all_colors = ["None", "white", "black", "pink", "red", "green", "yellow", "blue", "hex-color"]

        self.drop_width = drop_width
        self.on_choose_color = on_choose_color
        self.main_view = main_view
        self.selected_color = selected_color

        v = Container()
        self.v = v

        main_dropdown = flet.Dropdown(width=drop_width, value=selected_color, label=title_name,
                                      on_change=self.on_choose)
        for i in self.__all_colors:
            main_dropdown.options.append(flet.dropdown.Option(i))
        self.mainDropdown = main_dropdown

        color_preview = flet.Container(width=color_prev_width, height=50, bgcolor=self.selected_color, border_radius=8,
                                       border=flet.border.all(0.1, "white"))
        self.color_preview = color_preview

        main_row = Row(
            controls=[
                main_dropdown,
                color_preview
            ]
        )
        self.main_row = main_row

        v.content = main_row
        if add_it:
            main_view.controls.append(v)

    def on_choose(self, me):
        def on_change_color(me):
            tfc.value = str(tfc.value).replace(" ", "")
            tfc.update()
            if tfc.value != "#":
                new_color_selected = tfc.value
                self.selected_color = tfc.value
                self.color_preview.bgcolor = tfc.value
                self.on_choose_color(tfc.value)
            self.v.update()

        def back_to_orig(me):
            self.color_preview.on_click = None
            self.selected_color = "white"
            self.mainDropdown.value = self.selected_color
            self.color_preview.bgcolor = self.selected_color
            self.main_row.controls[0] = self.mainDropdown
            self.v.update()

        new_color_selected = me.control.value
        tfc = flet.TextField(
            label="hex-color",
            value="#0",
            width=self.drop_width,
            on_change=on_change_color,
            on_submit=on_change_color
        )

        if str(new_color_selected) == "hex-color":
            self.main_row.controls[0] = tfc
            self.color_preview.on_click = back_to_orig
        elif str(new_color_selected) == "None":
            self.color_preview.bgcolor = None
            self.selected_color = None
            self.on_choose_color(new_color_selected)
        else:
            self.color_preview.bgcolor = new_color_selected
            self.selected_color = new_color_selected
            self.on_choose_color(new_color_selected)

        self.v.update()

    def update(self):
        self.mainDropdown.value = self.selected_color
        self.color_preview.bgcolor = self.selected_color
        self.v.update()

    @property
    def value(self):
        return self.selected_color



import flet


class StoryBoard:
    def __init__(self, page, main_class):
        self.__page: flet.Page = page
        self.__main_class = main_class
        self.functions = {}

    def add_function(self, function_name, function):
        """
Define function, so the storyboard's widgets can access to. Like `function name` property of the `Button` widget need to have a Defined function.
        """
        self.functions[function_name] = function
        self.points = {}

    def add_flet_control(self, control, on_centered=True):
        """
        Add a flet control to the storyboard.
        if `on_centered = True` that means that the control will center the storyboard.
        """
        if on_centered:
            r = flet.Row([control], alignment=flet.MainAxisAlignment.CENTER)
            self.__page.add(r)
            self.__page.update()
        else:
            self.__page.add(control)

    def get_point(self, point_name: str):
        """
        The point is a place you choose to store some data in.
        You will get `None` as a return if the point have nothing yet.
        """
        if point_name not in self.points:
            return None
        else:
            return self.points[point_name]

    def navigate_to_page(self, page_name: str):
        """
        Go to a page name
        """
        if page_name not in self.__main_class.dict_content["pages"]:
            raise KeyError(f"Page '{page_name}' is not found.")

        viewerEngine = self.__main_class.viewerEngine
        viewerEngine = viewerEngine(
            self.__main_class,
            self.__main_class.dict_content,
            page_name,
            self.__page,
            self.__page,
            self.__main_class.development_mode
        )
        self.__page.update()

    def close_window(self):
        """Close the window"""
        self.__page.window_close()



import flet
from ..engines.viewer_engine import viewerEngine


class PreviewSection:
    def __init__(self, page: flet.Page, main_class, main_row: flet.Row):
        self.page: flet.Page = page
        self.main_class = main_class

        self.main_view = flet.Container(
            width=page.width - (page.width / 4) * 2,
            height=page.height - 80,
            border=flet.border.all(0.4, flet.colors.WHITE60),
            border_radius=12
        )
        main_row.controls.append(self.main_view)

        self.main_view_column = flet.Column(alignment=flet.MainAxisAlignment.CENTER)
        self.main_view.content = self.main_view_column

        self.update_preview()
        page.update()

    def update_preview(self, page_name="main"):
        ve = viewerEngine(
            self.main_class,
            self.main_class.dict_content,
            page_name,
            self.main_view,
            self.main_view_column
        )



from setuptools import setup, find_packages

setup(
    name='Flet StoryBoard',
    version='1.3',
    author='SKbarbon',
    description='A UI-Builder that helps programmers build the front-end without codding it.',
    long_description="https://github.com/SKbarbon/Flet_StoryBoard",
    url='https://github.com/SKbarbon/Flet-StoryBoard',
    install_requires=["flet==0.5.2", "requests"],
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows"
    ],
)


import flet
from ...tools.color_picker import ColorPicker
from ...tools.list_picker import ListPopup


class Open_Url(object):
    def __init__(self, main_class, parent, *args, **kwargs) -> None:
        self.parent = parent
        self.main_class = main_class
        self.self_object = flet.Container(on_click=self.open_url)
        self.url_text = flet.Text()
        self.url_icon = flet.Icon(size=12)
        self.self_object.content = flet.Row([self.url_icon, self.url_text], alignment=flet.MainAxisAlignment.CENTER)

        # all args
        all_icons = []
        for icon in flet.icons.__dict__:
            name = str(icon)
            if name.startswith("__"):
                pass
            else:
                all_icons.append(name)

        self.args = {
            "text": {"type": str, "default_value": "example url"},
            "url": {"type": str, "default_value": "https://example.com"},
            "link icon": {"type": ListPopup, "options": all_icons, "default_value": "INSERT_LINK_OUTLINED"},
            "text_color": {"type": ColorPicker, "default_value": "blue"},
            "icon_color": {"type": ColorPicker, "default_value": "blue"},
            "bgcolor": {"type": ColorPicker, "default_value": "black"},
            "alignment": {"type": list, "options": ["left", "center", "right"], "default_value": "center"},
            "width": {"type": int, "default_value": 130},
            "height": {"type": int, "default_value": 30},
            "text_size": {"type": int, "default_value": 15},
            "icon_size": {"type": int, "default_value": 15},
            "border_radius": {"type": int, "default_value": 10},
            "hide": {"type": bool, "default_value": False}
        }

        # Template dict
        # This is where the widget data will be stored.
        self.template = {
            "widget_class_name": "Open Url",
            "properties": {}
        }
        for p in self.args:
            self.template["properties"][p] = self.args[p]["default_value"]

        self.update()

    def update(self, new_props: dict = None):
        c = self.self_object
        t = self.url_text
        ii = self.url_icon
        props = self.template["properties"]

        if new_props is not None:
            for i in new_props:
                self.template["properties"][i] = new_props[i]

        url = props["url"]
        c.tooltip = f"open '{url}'"
        c.width = props["width"]
        c.bgcolor = props["bgcolor"]
        c.border_radius = props["border_radius"]
        c.height = props["height"]
        c.visible = props["hide"] == False

        t.value = props["text"]
        t.color = props["text_color"]
        t.text_align = props["text_color"]
        t.size = props["text_size"]

        ii.name = props["link icon"]
        ii.color = props["icon_color"]
        ii.size = props["icon_size"]
        if self.self_object.page is not None:
            self.self_object.update()

    def open_url(self, event):
        if self.main_class.development_mode: return
        url = self.template["properties"]["url"]
        event.control.page.launch_url(url)

    def return_widget(self):
        props = self.template["properties"]
        if props["alignment"] == "left":
            return flet.Row([flet.Text("   "), self.self_object])
        elif props["alignment"] == "center":
            return flet.Row([self.self_object], alignment=flet.MainAxisAlignment.CENTER)
        else:
            # issue: right is not a valid value
            return flet.Row([self.self_object, flet.Text("    ")], alignment="right")



from .pages.create_new_file import CreateNewFile
from .pages.main_page import mainPage
import sys
import os
import flet

cmd_args = sys.argv
debug_mode = False
if "--debug" in cmd_args:
    debug_mode = True


class manage_edit:
    def __init__(self) -> None:
        self.cmd_args = sys.argv
        self.file_name = None
        flet.app(target=self.CNF)

        if self.file_name is None and debug_mode:
            print("Debug alert: Unexpected exit, no errors.")
            sys.exit("Exit.")

        if not str(self.file_name).endswith(".fletsb"):
            self.file_name = str(self.file_name) + ".fletsb"

        if debug_mode:
            print("Debug alert: About to run the editor.")

        mainPage(self.file_name)

    def CNF(self, page):
        CreateNewFile(page, manage_class=self)


if len(cmd_args) == 1:
    # To create a new file.
    if debug_mode:
        print("Debug alert: About to show the 'CreateNewFile' window.")
    manage_edit()
else:
    # To edit an existing File.
    full_path = ""
    for i in range(1, len(cmd_args)):
        if full_path == "":
            full_path = full_path + cmd_args[i]
        else:
            full_path = full_path + " " + cmd_args[i]

    full_path = full_path.replace(" --debug", "")
    full_path = full_path.replace("--debug", "")

    if os.path.isfile(full_path):
        if debug_mode:
            print("Debug alert: About to run the editor.")
        mainPage(full_path)
    else:
        print(f"Warning: File not found '{full_path}', so create a new one.")
        if debug_mode:
            print("Debug alert: About to show the 'CreateNewFile' window.")
        manage_edit()
